from Products.CMFCore.utils import getToolByName
from ftw.upgrade.workflow import WorkflowChainUpdater
import logging


LOG = logging.getLogger('ftw.upgrade.PlacefulWorkflowPolicyActivator')


class PlacefulWorkflowPolicyActivator(object):

    def __init__(self, context):
        self.context = context

    def activate_policy(self, policy_id, review_state_mapping,
                        activate_in=True, activate_below=True,
                        **chain_updater_kwargs):

        with WorkflowChainUpdater(self.get_objects(), review_state_mapping,
                                  **chain_updater_kwargs):
            self._activate_placeful_policy(policy_id,
                                           activate_in=activate_in,
                                           activate_below=activate_below)

    def get_objects(self):
        objects = []

        def recurse(obj):
            objects.append(obj)
            for child in obj.objectValues():
                recurse(child)

        recurse(self.context)
        return objects

    def _activate_placeful_policy(self, policy_id,
                                  activate_in=True, activate_below=True):
        LOG.info('Activating placeful policy %s' % policy_id)
        pwf_tool = getToolByName(self.context, 'portal_placeful_workflow')
        policy_config = pwf_tool.getWorkflowPolicyConfig(self.context)

        if not policy_config:
            self.context.manage_addProduct[
                'CMFPlacefulWorkflow'].manage_addWorkflowPolicyConfig()
            policy_config = pwf_tool.getWorkflowPolicyConfig(self.context)

        if activate_in:
            policy_config.setPolicyIn(policy_id, update_security=False)

        if activate_below:
            policy_config.setPolicyBelow(policy_id, update_security=False)



from collections import defaultdict
from copy import deepcopy
from ftw.upgrade.exceptions import CyclicDependencies
from path import Path
import logging
import math
import os
import stat
import tarjan.tc
import transaction


def topological_sort(items, partial_order):
    """Perform topological sort.
    items is a list of items to be sorted.
    partial_order is a list of pairs. If pair (a,b) is in it, it means
    that item a should appear before item b.
    Returns a list of the items in one of the possible orders, or None
    if partial_order contains a loop.
    Source: http://www.bitformation.com/art/python_toposort.html
    """

    def add_node(graph, node):
        """Add a node to the graph if not already exists."""
        if node not in graph:
            graph[node] = [0]  # 0 = number of arcs coming into this node.

    def add_arc(graph, fromnode, tonode):
        """Add an arc to a graph. Can create multiple arcs.
        The end nodes must already exist."""
        graph[fromnode].append(tonode)
        # Update the count of incoming arcs in tonode.
        graph[tonode][0] = graph[tonode][0] + 1

    # step 1 - create a directed graph with an arc a->b for each input
    # pair (a,b).
    # The graph is represented by a dictionary. The dictionary contains
    # a pair item:list for each node in the graph. /item/ is the value
    # of the node. /list/'s 1st item is the count of incoming arcs, and
    # the rest are the destinations of the outgoing arcs. For example:
    #           {'a':[0,'b','c'], 'b':[1], 'c':[1]}
    # represents the graph:   c <-- a --> b
    # The graph may contain loops and multiple arcs.
    # Note that our representation does not contain reference loops to
    # cause GC problems even when the represented graph contains loops,
    # because we keep the node names rather than references to the nodes.
    graph = {}
    for v in items:
        add_node(graph, v)
    for a, b in partial_order:
        add_arc(graph, a, b)

    # Step 2 - find all roots (nodes with zero incoming arcs).
    roots = [node for (node, nodeinfo) in graph.items() if nodeinfo[0] == 0]
    roots.sort()

    # step 3 - repeatedly emit a root and remove it from the graph. Removing
    # a node may convert some of the node's direct children into roots.
    # Whenever that happens, we append the new roots to the list of
    # current roots.
    sorted_ = []
    while len(roots) != 0:
        # If len(roots) is always 1 when we get here, it means that
        # the input describes a complete ordering and there is only
        # one possible output.
        # When len(roots) > 1, we can choose any root to send to the
        # output; this freedom represents the multiple complete orderings
        # that satisfy the input restrictions. We arbitrarily take one of
        # the roots using pop(). Note that for the algorithm to be efficient,
        # this operation must be done in O(1) time.
        root = roots.pop()
        sorted_.append(root)
        for child in graph[root][1:]:
            graph[child][0] = graph[child][0] - 1
            if graph[child][0] == 0:
                roots.append(child)
        del graph[root]
    if len(graph.items()) != 0:
        # There is a loop in the input.
        return None
    return sorted_


def find_cyclic_dependencies(dependencies):
    deps = defaultdict(list)
    for first, second in dependencies:
        deps[first].append(second)

    cyclic_dependencies = []
    for name, closure in tarjan.tc.tc(deps).items():
        if name in closure and closure not in cyclic_dependencies:
            cyclic_dependencies.append(closure)

    return cyclic_dependencies


class SizedGenerator(object):

    def __init__(self, generator, length):
        self._length = length
        self._generator = generator

    def __iter__(self):
        return self._generator.__iter__()

    def __len__(self):
        return self._length


class SavepointIterator(object):
    """An iterator that creates a savepoint every n items.
    """

    def __init__(self, iterable, threshold, logger=None):
        self.iterable = iterable
        self.threshold = threshold
        self.logger = logger

        if self.logger is None:
            self.logger = logging.getLogger('ftw.upgrade')

        if not threshold:
            raise ValueError("Threshold must be a non-zero value")

    def __iter__(self):
        for i, item in enumerate(self.iterable):
            if i % self.threshold == 0:
                transaction.savepoint()
                self.logger.info("Created savepoint at %s items" % i)
            yield item

    def __len__(self):
        return self.iterable.__len__()

    @classmethod
    def build(cls, iterable, threshold=None, logger=None):
        if threshold:
            return SavepointIterator(iterable, threshold, logger)
        else:
            return iterable


def get_sorted_profile_ids(portal_setup):
    """Returns a sorted list of profile ids (without profile- prefix).
    The sorting is done by resolving the dependencies and performing
    a topological graph sort.
    If there are circular dependencies a CyclicDependencies exception
    is thrown.
    """
    profile_ids = []
    dependencies = []

    for profile in portal_setup.listProfileInfo():
        profile_ids.append(profile['id'])

    for profile in portal_setup.listProfileInfo():
        if not profile.get('dependencies'):
            continue

        for dependency in profile.get('dependencies'):
            if dependency.startswith('profile-'):
                dependency = dependency.split('profile-', 1)[1]
            else:
                continue

            if dependency not in profile_ids:
                continue
            dependencies.append((profile['id'], dependency))

    order = topological_sort(profile_ids, dependencies)

    if order is None:
        raise CyclicDependencies(
            dependencies,
            find_cyclic_dependencies(deepcopy(dependencies)))
    else:
        return list(reversed(order))


def format_duration(seconds):
    """Makes a duration in seconds human readable.
    Supports hours, minutes and seconds.
    """

    seconds = math.ceil(seconds)
    hours, remainder = divmod(seconds, 60 * 60)
    minutes, seconds = divmod(remainder, 60)

    result = []

    if hours == 1:
        result.append('1 hour')
    elif hours > 1:
        result.append('%i hours' % hours)

    if minutes == 1:
        result.append('1 minute')
    elif minutes > 1:
        result.append('%i minutes' % minutes)

    if seconds == 1:
        result.append('1 second')
    elif seconds > 1:
        result.append('%i seconds' % seconds)

    if len(result) == 0:
        return '0 seconds'
    else:
        return ', '.join(result)


def subject_from_docstring(docstring):
    """Extracts and returns the subject of a docstring.
    The subject consists of all lines from the beginning to the
    first empty line. Newlines are stripped.
    """
    lines = map(str.strip, docstring.strip().splitlines())
    try:
        lines.index('')
    except ValueError:
        pass  # '' is not in list
    else:
        lines = lines[:lines.index('')]

    return ' '.join(lines).strip()


def get_tempfile_authentication_directory(directory=None):
    """Finds the buildout directory and returns the absolute path to the
    relative directory var/ftw.upgrade-authentication/.
    If the directory does not exist it is created.
    """
    directory = Path(directory) or Path.getcwd()
    if not directory.joinpath('bin', 'buildout').isfile():
        return get_tempfile_authentication_directory(directory.parent)

    auth_directory = directory.joinpath('var', 'ftw.upgrade-authentication')
    if not auth_directory.isdir():
        auth_directory.mkdir(mode=0700)

    if stat.S_IMODE(auth_directory.stat().st_mode) != 0700:
        raise ValueError('{0} has invalid mode.'.format(auth_directory))
    if auth_directory.stat().st_uid != os.getuid():
        raise ValueError('{0} has an invalid owner.'.format(auth_directory))

    return auth_directory



from contextlib import closing
from ftw.upgrade.command.jsonapi import add_requestor_authentication_argument
from ftw.upgrade.command.jsonapi import add_site_path_argument
from ftw.upgrade.command.jsonapi import error_handling
from ftw.upgrade.command.jsonapi import with_api_requestor
from ftw.upgrade.command.terminal import TERMINAL
import re
import sys


DOCS = """
{t.bold}DESCRIPTION:{t.normal}
    Install upgrades or profiles. Either the "--proposed" or the \
    "--upgrades" or the "--profiles" argument must be used.

{t.bold}PROPOSED UPGRADES:{t.normal}
    When using the "--proposed" argument, all proposed upgrade steps are\
 installed, ordered by dependencies and versions.

{t.bold}SET OF UPGRADES:{t.normal}
    By using the "--upgrades" argument, one or many upgrades can be \
installed. \
Each upgrade is identified by its API id of format "<dest>@<profileid>".

{t.bold}INSTALL PROFILES:{t.normal}
    By using the "--profiles" argument, one or many profiles can be \
    installed. Each install is identified by profile id.

{t.bold}INSTALL ORDER:{t.normal}
    The upgrades are always reordered before they are installed. \
The profiles are ordered topologically with the GS profile dependency graph. \
Upgrades within each profile are ordered by the source / destination versions.

{t.bold}EXAMPLES:{t.normal}
[quote]
$ bin/upgrade install --site Plone --proposed
$ bin/upgrade install --site Plone --proposed --auth admin:admin
$ bin/upgrade install --site Plone --upgrades 3001@my.package:default \
3002@my.package:default
$ bin/upgrade install --site Plone --profiles Products.PloneFormGen:default \
unwanted.addon:uninstall
[/quote]

""".format(t=TERMINAL).strip()


def valid_upgrade_step_id(value):
    if not re.match(r'^\w+@[\w\.]+:\w+$', value):
        raise ValueError(value)
    return value


def valid_profile_id(value):
    if ':' not in value:
        raise ValueError(value)
    return value


def setup_argparser(commands):
    command = commands.add_parser(
        'install',
        help='Install upgrades or profiles.',
        description=DOCS)
    command.set_defaults(func=install_command)
    add_requestor_authentication_argument(command)
    add_site_path_argument(command)

    group = command.add_mutually_exclusive_group(required=True)
    group.add_argument('--upgrades', '-u', nargs='+',
                       help='One or many upgrade step API ids.',
                       type=valid_upgrade_step_id)
    group.add_argument('--proposed', '-p', action='store_true',
                       help='Installs all proposed upgrades.')
    group.add_argument('--profiles', nargs='+',
                       help='One or many profile ids.',
                       type=valid_profile_id)


@with_api_requestor
@error_handling
def install_command(args, requestor):
    if args.proposed:
        action = 'execute_proposed_upgrades'
        params = ()
    elif args.profiles:
        action = 'execute_profiles'
        params = [('profiles:list', name) for name in set(args.profiles)]
    else:
        action = 'execute_upgrades'
        params = [('upgrades:list', name) for name in set(args.upgrades)]

    with closing(requestor.POST(action, params=params,
                                stream=True)) as response:
        for line in response.iter_lines(chunk_size=30, decode_unicode=True):
            print line

    if line.strip() != 'Result: SUCCESS':
        sys.exit(3)



import transaction


TRANSACTION_NOTE_MAX_LENGTH = 65533


class TransactionNote(object):
    """The zope transaction note is limited to a length of 60000 characters.
    When installing a lot of upgrades at once, this limitation may be
    exceeded.

    Transaction not strategy:
    - Use "profileid -> dest (description)" for each upgrade
    - when too long, use "profileid -> dest" (without the description)
    - when still too long, crop at the end, appending "..."
    """

    def add_upgrade(self, profileid, destination, description):
        self._upgrades.append({'profileid': profileid,
                               'destination': '.'.join(destination),
                               'description': description})

    def set_transaction_note(self):
        maximum_possible_length = TRANSACTION_NOTE_MAX_LENGTH \
            - len(transaction.get().description) \
            - len('\n')

        message = '\n'.join(self._transaction_messages(True))
        if len(message) >= maximum_possible_length:
            message = '\n'.join(self._transaction_messages(False))

        if len(message) >= maximum_possible_length:
            message = message[:maximum_possible_length - 4] + '...'

        if len(message) <= maximum_possible_length:
            transaction.get().note(message)

    def _transaction_messages(self, include_description=True):
        if include_description:
            template = '%(profileid)s -> %(destination)s (%(description)s)'
        else:
            template = '%(profileid)s -> %(destination)s'

        return [template % upgrade for upgrade in self._upgrades]

    def _reset_upgrade_info(self):
        current_transaction = transaction.get()
        if not hasattr(current_transaction, 'ftw.upgrade:upgrades'):
            setattr(current_transaction, 'ftw.upgrade:upgrades', [])

    @property
    def _upgrades(self):
        self._reset_upgrade_info()
        current_transaction = transaction.get()
        return getattr(current_transaction, 'ftw.upgrade:upgrades')



from DateTime import DateTime
from datetime import datetime
from ftw.builder import Builder
from ftw.builder import create
from ftw.upgrade import UpgradeStep
from ftw.upgrade.exceptions import NoAssociatedProfileError
from ftw.upgrade.interfaces import IUpgradeStep
from ftw.upgrade.tests.base import UpgradeTestCase
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.browserlayer.utils import register_layer
from Products.CMFCore.utils import getToolByName
from StringIO import StringIO
from zope.interface import Interface
from zope.interface.verify import verifyClass
import logging


class IMyProductLayer(Interface):
    """Dummy class used in test_remove_broken_browserlayer()
    """


class TestUpgradeStep(UpgradeTestCase):

    def setUp(self):
        super(TestUpgradeStep, self).setUp()
        self.portal = self.layer['portal']
        self.portal_setup = getToolByName(self.portal, 'portal_setup')

        self.log = StringIO()
        self.logger = logging.getLogger('ftw.upgrade')
        self.logger.setLevel(logging.DEBUG)

        handler = logging.StreamHandler(self.log)
        self.logger.addHandler(handler)

        setRoles(self.portal, TEST_USER_ID, ['Manager'])

    def test_implements_interface(self):
        verifyClass(IUpgradeStep, UpgradeStep)

    def test_upgrade_classmethod(self):
        with self.assertRaises(NotImplementedError):
            UpgradeStep(self.portal_setup)

    def test_portal_setup_attribute(self):
        testcase = self

        class Step(UpgradeStep):
            def __call__(self):
                testcase.assertEqual(testcase.portal_setup, self.portal_setup)

        Step(self.portal_setup)

    def test_portal_attribute(self):
        testcase = self

        class Step(UpgradeStep):
            def __call__(self):
                testcase.assertEqual(testcase.portal, self.portal)

        Step(self.portal_setup)

    def test_getToolByName(self):
        actions_tool = getToolByName(self.portal, 'portal_actions')
        testcase = self

        class Step(UpgradeStep):
            def __call__(self):
                testcase.assertEqual(
                    actions_tool,
                    self.getToolByName('portal_actions'))

        Step(self.portal_setup)

    def test_objects_method_yields_objects_with_logging(self):
        testcase = self
        create(Builder('folder').titled('Foo'))
        create(Builder('folder').titled('Bar'))
        create(Builder('folder').titled('Baz'))

        object_titles = []

        class Step(UpgradeStep):
            def __call__(self):
                for obj in self.objects({'portal_type': 'Folder'},
                                        'Log message',
                                        logger=testcase.logger):
                    object_titles.append(obj.Title())

        Step(self.portal_setup)

        self.assertEquals(set(['Foo', 'Bar', 'Baz']), set(object_titles))

        self.assertEquals(['STARTING Log message',
                           '1 of 3 (33%): Log message',
                           'DONE Log message'],
                          self.read_log())

    def test_objects_modifying_catalog_does_not_reduce_result_set(self):
        old_date = DateTime(2010, 1, 1)
        new_date = DateTime(2012, 3, 3)
        create(Builder('folder').with_modification_date(old_date))
        create(Builder('folder').with_modification_date(old_date))
        create(Builder('folder').with_modification_date(old_date))
        create(Builder('folder').with_modification_date(old_date))

        data = {'processed_folders': 0}

        class Step(UpgradeStep):
            def __call__(self):
                for obj in self.objects({'modified': old_date}, ''):
                    obj.setModificationDate(new_date)
                    obj.reindexObject(idxs=['modified'])
                    data['processed_folders'] += 1

        Step(self.portal_setup)
        self.assertEquals(
            4, data['processed_folders'],
            'Updating catalog reduced result set while iterating over it!!!')

    def test_catalog_rebuild_index(self):
        testcase = self

        class Step(UpgradeStep):
            def __call__(self):
                ctool = self.getToolByName('portal_catalog')
                name = 'getExcludeFromNav'

                self.catalog_add_index(name, 'BooleanIndex')
                testcase.assertEqual(0, len(ctool._catalog.getIndex(name)))
                self.catalog_rebuild_index(name)
                testcase.assertEqual(1, len(ctool._catalog.getIndex(name)))

        create(Builder('folder')
               .titled('Rebuild Index Test Obj')
               .having(excludeFromNav=True))

        Step(self.portal_setup)

    def test_catalog_reindex_objects(self):
        testcase = self
        create(Builder('folder'))

        class Step(UpgradeStep):
            def __call__(self):
                ctool = self.getToolByName('portal_catalog')
                name = 'getExcludeFromNav'

                self.catalog_add_index(name, 'BooleanIndex')
                testcase.assertEqual(0, len(ctool._catalog.getIndex(name)))

                self.catalog_reindex_objects({'portal_type': 'Folder'})

                self.catalog_rebuild_index(name)
                testcase.assertEqual(1, len(ctool._catalog.getIndex(name)))

        Step(self.portal_setup)

    def test_catalog_reindex_objects_keeps_modification_date(self):
        testcase = self
        folder = create(Builder('folder'))

        modification_date = DateTime('2010/12/31')
        folder.setModificationDate(modification_date)
        folder.reindexObject(idxs=['modified'])

        class Step(UpgradeStep):
            def __call__(self):

                brain = self.catalog_unrestricted_search(
                    {'UID': folder.UID()})[0]
                testcase.assertEquals(modification_date, brain.modified)

                self.catalog_reindex_objects({})

                brain = self.catalog_unrestricted_search(
                    {'UID': folder.UID()})[0]
                testcase.assertEquals(modification_date, brain.modified)

        Step(self.portal_setup)

    def test_catalog_has_index(self):
        testcase = self

        class Step(UpgradeStep):
            def __call__(self):
                testcase.assertTrue(self.catalog_has_index('sortable_title'))
                testcase.assertFalse(self.catalog_has_index('foo'))

        Step(self.portal_setup)

    def test_catalog_add_and_remove_indexes(self):
        testcase = self

        class Step(UpgradeStep):
            def __call__(self):
                testcase.assertFalse(self.catalog_has_index('foo'))
                self.catalog_add_index('foo', 'FieldIndex')
                testcase.assertTrue(self.catalog_has_index('foo'))
                self.catalog_remove_index('foo')
                testcase.assertFalse(self.catalog_has_index('foo'))

        Step(self.portal_setup)

    def test_catalog_unrestricted_get_object(self):
        testcase = self
        folder = create(Builder('folder'))
        create(Builder('document').within(folder))

        folder_path = '/'.join(folder.getPhysicalPath())

        class Step(UpgradeStep):
            def __call__(self):
                query = {'path': folder_path,
                         'portal_type': 'Document'}
                brains = self.catalog_unrestricted_search(query)

                brain = brains[0]
                testcase.assertEqual(
                    'ImplicitAcquisitionWrapper',
                    type(brain).__name__)

                obj = self.catalog_unrestricted_get_object(brain)
                testcase.assertEqual(
                    'ImplicitAcquisitionWrapper',
                    type(obj).__name__)

        Step(self.portal_setup)

    def test_catalog_unrestricted_search(self):
        testcase = self

        folder = create(Builder('folder'))
        create(Builder('document').titled('Page One').within(folder))
        create(Builder('document').titled('Page Two').within(folder))

        folder_path = '/'.join(folder.getPhysicalPath())

        class Step(UpgradeStep):
            def __call__(self):
                query = {'path': folder_path,
                         'portal_type': 'Document'}
                brains = self.catalog_unrestricted_search(query)

                testcase.assertEqual(2, len(brains))
                testcase.assertEqual(['page-one', 'page-two'],
                                     [brain.id for brain in brains])
                testcase.assertEqual('ImplicitAcquisitionWrapper',
                                     type(brains[0]).__name__)

                objects = self.catalog_unrestricted_search(
                    query, full_objects=True)
                testcase.assertEqual(2, len(objects))
                objects = list(objects)
                testcase.assertEqual(['page-one', 'page-two'],
                                     [obj.id for obj in objects])
                testcase.assertEqual('ImplicitAcquisitionWrapper',
                                     type(objects[0]).__name__)

        Step(self.portal_setup)

    def test_actions_remove_action(self):
        atool = getToolByName(self.portal, 'portal_actions')
        self.assertIn('rss', atool.get('document_actions'))
        testcase = self

        class Step(UpgradeStep):
            def __call__(self):
                testcase.assertTrue(self.actions_remove_action(
                        'document_actions', 'rss'))
                testcase.assertFalse(self.actions_remove_action(
                        'document_actions', 'rss'))

        Step(self.portal_setup)
        self.assertNotIn('rss', atool.get('document_actions'))

    def test_actions_remove_type_action(self):
        testcase = self

        class Step(UpgradeStep):
            def event_has_action(self, actionid):
                ttool = self.getToolByName('portal_types')
                fti = ttool.get('Event')

                for action in fti._actions:
                    if action.id == actionid:
                        return True

                return False

            def __call__(self):
                testcase.assertTrue(self.event_has_action('history'))
                testcase.assertTrue(
                    self.actions_remove_type_action('Event', 'history'))
                testcase.assertFalse(self.event_has_action('history'))

        Step(self.portal_setup)

    def test_actions_add_type_action(self):
        testcase = self

        class Step(UpgradeStep):
            def get_event_action(self, actionid):
                ttool = self.getToolByName('portal_types')
                fti = ttool.get('Event')

                for action in fti._actions:
                    if action.id == actionid:
                        return action

            def get_action_ids(self):
                ttool = self.getToolByName('portal_types')
                fti = ttool.get('Event')

                return [action.id for action in fti._actions]

            def __call__(self):
                testcase.assertEquals(
                    None, self.get_event_action('additional'))
                testcase.assertEquals(
                    ['view', 'edit', 'history', 'external_edit'],
                    self.get_action_ids())

                self.actions_add_type_action(
                    'Event', 'history', action_id='additional', title='Additional',
                    action='string:#', permissions=('View', ))

                testcase.assertEquals(
                    ['view', 'edit', 'history', 'additional', 'external_edit'],
                    self.get_action_ids())

                action = self.get_event_action('additional')
                testcase.assertEquals('Additional', action.title)
                testcase.assertEquals('string:#', action.action.text)

        Step(self.portal_setup)

    def test_set_property(self):
        testcase = self

        class Step(UpgradeStep):
            def __call__(self):
                testcase.assertFalse(self.portal.hasProperty('foo'))

                self.set_property(self.portal, 'foo', 'bar')
                testcase.assertTrue(self.portal.hasProperty('foo'))
                testcase.assertEqual('bar', self.portal.getProperty('foo'))

                self.set_property(self.portal, 'foo', 'baz')
                testcase.assertEqual('baz', self.portal.getProperty('foo'))

        Step(self.portal_setup)

    def test_add_lines_to_property(self):
        testcase = self

        class Step(UpgradeStep):
            def __call__(self):
                key = 'foo-lines'

                testcase.assertFalse(self.portal.hasProperty(key))

                self.add_lines_to_property(self.portal, key, 'foo')
                testcase.assertEqual(('foo',),
                                     self.portal.getProperty(key))

                self.add_lines_to_property(self.portal, key, ['bar'])
                testcase.assertEqual(('foo', 'bar'),
                                     self.portal.getProperty(key))

                self.add_lines_to_property(self.portal, key, 'baz')
                testcase.assertEqual(('foo', 'bar', 'baz'),
                                     self.portal.getProperty(key))

                self.set_property(self.portal, key, ('foo'))
                testcase.assertEqual(('foo',),
                                     self.portal.getProperty(key))

                self.add_lines_to_property(self.portal, key, ['bar'])
                testcase.assertEqual(('foo', 'bar'),
                                     self.portal.getProperty(key))

                self.add_lines_to_property(self.portal, key, ('baz',))
                testcase.assertEqual(('foo', 'bar', 'baz'),
                                     self.portal.getProperty(key))

                self.portal.manage_delProperties([key])
                self.add_lines_to_property(self.portal, key, ['foo', 'bar'])
                testcase.assertEqual(('foo', 'bar'),
                                     self.portal.getProperty(key))

        Step(self.portal_setup)

    def test_setup_install_profile(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_file('catalog.xml', self.asset('exclude-from-nav-index.xml')))

        testcase = self

        class Step(UpgradeStep):
            def __call__(self):
                testcase.assertFalse(self.catalog_has_index('excludeFromNav'))
                self.setup_install_profile('profile-the.package:default')
                testcase.assertTrue(self.catalog_has_index('excludeFromNav'))

                self.catalog_remove_index('excludeFromNav')
                testcase.assertFalse(self.catalog_has_index('excludeFromNav'))
                self.setup_install_profile('profile-the.package:default', ['catalog'])
                testcase.assertTrue(self.catalog_has_index('excludeFromNav'))

        with self.package_created():
            Step(self.portal_setup)

    def test_install_upgrade_profile(self):
        class AddSiteProperty(UpgradeStep):
            def __call__(self):
                assert not self.portal.getProperty('foo'), \
                    'property "foo" should not yet exist'
                self.install_upgrade_profile()
                assert self.portal.getProperty('foo') == 'bar', \
                    'property "foo" was not created or is incorrect.'

        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step')
                          .to(datetime(2011, 1, 1))
                          .calling(AddSiteProperty)
                          .with_file('properties.xml', self.asset('foo-property.xml'))))

        with self.package_created():
            self.install_profile('the.package:default', '0')
            self.install_profile_upgrades('the.package:default')

    def test_install_upgrade_profile_raises_exception_when_no_profile_defined(self):
        class Step(UpgradeStep):
            def __call__(self):
                self.install_upgrade_profile()

        with self.assertRaises(NoAssociatedProfileError):
            Step(self.portal_setup)

    def test_uninstall_product(self):
        quickinstaller = getToolByName(self.portal, 'portal_quickinstaller')
        quickinstaller.installProduct('CMFPlacefulWorkflow')

        def installed_products():
            for product in quickinstaller.listInstalledProducts():
                yield product['id']

        class Step(UpgradeStep):
            def __call__(self):
                self.uninstall_product('CMFPlacefulWorkflow')

        self.assertIn('CMFPlacefulWorkflow', installed_products())
        Step(self.portal_setup)
        self.assertNotIn('CMFPlacefulWorkflow', installed_products())

    def test_migrate_class(self):
        from Products.ATContentTypes.content.folder import ATBTreeFolder

        folder = create(Builder('folder'))
        subfolder = create(Builder('folder').within(folder))


        class Step(UpgradeStep):
            def __call__(self):
                self.migrate_class(subfolder, ATBTreeFolder)

        self.assertEqual('ATFolder', subfolder.__class__.__name__)
        Step(self.portal_setup)
        self.assertEqual('ATBTreeFolder', subfolder.__class__.__name__)

    def test_migrate_class_also_updates_provided_interfaces_info(self):
        from Products.ATContentTypes.content.link import ATLink
        from Products.ATContentTypes.interfaces import IATLink
        from Products.ATContentTypes.interfaces import IATDocument

        obj = create(Builder('document'))
        self.assertTrue(IATDocument.providedBy(obj))
        self.assertFalse(IATLink.providedBy(obj))

        class Step(UpgradeStep):
            def __call__(self):
                self.migrate_class(obj, ATLink)

        Step(self.portal_setup)
        self.assertFalse(IATDocument.providedBy(obj),
                         'IATDocument interface not removed in migration')
        self.assertTrue(IATLink.providedBy(obj),
                        'IATLink interface not added in migration')

    def test_remove_broken_browserlayer(self):
        from plone.browserlayer.utils import registered_layers
        from plone.browserlayer.interfaces import ILocalBrowserLayerType
        register_layer(IMyProductLayer, 'my.product')

        class Step(UpgradeStep):
            def __call__(self):
                self.remove_broken_browserlayer('my.product',
                                                'IMyProductLayer')
        Step(self.portal_setup)

        # Check that it worked.  Note that this check fails if you have called
        # registered_layers earlier in this test before calling
        # remove_broken_browserlayer: the previous answer is cached.
        self.assertFalse(IMyProductLayer in registered_layers())
        # Check it a bit more low level.
        sm = self.portal.getSiteManager()
        adapters = sm.utilities._adapters
        self.assertFalse('my.product' in adapters[0][ILocalBrowserLayerType])
        subscribers = sm.utilities._subscribers
        layer_subscribers = subscribers[0][ILocalBrowserLayerType]
        iface_name = 'IMyProductLayer'
        self.assertEqual(len([layer for layer in layer_subscribers['']
                              if layer.__name__ == iface_name]), 0)


    def test_update_security_removes_roles_unmanaged_by_workflow(self):
        self.set_workflow_chain(for_type='Folder',
                                to_workflow='plone_workflow')
        folder = create(Builder('folder'))

        self.assert_permission_not_acquired('View', folder)
        self.assert_permission_not_acquired('Modify portal content', folder)
        self.assert_permission_acquired('List folder contents', folder)

        folder.manage_permission('Modify portal content', roles=[],
                                 acquire=True)
        folder.manage_permission('List folder contents', roles=[],
                                 acquire=False)

        self.assert_permission_not_acquired('View', folder)
        self.assert_permission_acquired('Modify portal content', folder)
        self.assert_permission_not_acquired('List folder contents', folder)

        class Step(UpgradeStep):
            def __call__(self):
                self.update_security(folder)
        Step(self.portal_setup)

        self.assert_permission_not_acquired('View', folder)
        self.assert_permission_not_acquired('Modify portal content', folder)
        self.assert_permission_acquired('List folder contents', folder)

    def test_update_security_reindexes_security_indexes(self):
        self.set_workflow_chain(for_type='Folder',
                                to_workflow='simple_publication_workflow')
        folder = create(Builder('folder')
                        .in_state('published'))

        self.assertEquals(['Anonymous'],
                          self.get_allowed_roles_and_users_for(folder))
        folder.reindexObjectSecurity()

        folder.manage_permission('View', roles=['Reader'], acquire=False)
        folder.reindexObjectSecurity()

        self.assertEquals(['Reader'],
                          self.get_allowed_roles_and_users_for(folder))

        class Step(UpgradeStep):
            def __call__(self):
                self.update_security(folder)
        Step(self.portal_setup)

        self.assertEquals(['Anonymous'],
                          self.get_allowed_roles_and_users_for(folder))

    def test_update_security_without_reindexing_security(self):
        self.set_workflow_chain(for_type='Folder',
                                to_workflow='simple_publication_workflow')
        folder = create(Builder('folder')
                        .in_state('published'))

        self.assertEquals(['Anonymous'],
                          self.get_allowed_roles_and_users_for(folder))

        folder.manage_permission('View', roles=['Reader'], acquire=False)
        folder.reindexObjectSecurity()

        self.assertEquals(['Reader'],
                          self.get_allowed_roles_and_users_for(folder))

        class Step(UpgradeStep):
            def __call__(self):
                self.update_security(folder, reindex_security=False)
        Step(self.portal_setup)

        self.assertEquals(['Reader'],
                          self.get_allowed_roles_and_users_for(folder))

    def test_update_workflow_security_updates_security(self):
        self.set_workflow_chain(for_type='Folder',
                                to_workflow='plone_workflow')
        folder = create(Builder('folder'))
        self.assert_permission_not_acquired('Modify portal content', folder)
        folder.manage_permission('Modify portal content', roles=[],
                                 acquire=True)
        self.assert_permission_acquired('Modify portal content', folder)

        class Step(UpgradeStep):
            def __call__(self):
                self.update_workflow_security(['plone_workflow'])
        Step(self.portal_setup)

        self.assert_permission_not_acquired('Modify portal content', folder)

    def test_update_workflow_security_reindexes_security(self):
        self.set_workflow_chain(for_type='Folder',
                                to_workflow='plone_workflow')
        folder = create(Builder('folder').in_state('published'))

        folder.manage_permission('View', roles=['Reader'], acquire=False)
        folder.reindexObjectSecurity()
        self.assertEquals(['Reader'],
                          self.get_allowed_roles_and_users_for(folder))

        class Step(UpgradeStep):
            def __call__(self):
                self.update_workflow_security(
                    ['plone_workflow'], reindex_security=False)
        Step(self.portal_setup)

        self.assertEquals(['Reader'],
                          self.get_allowed_roles_and_users_for(folder))

        class Step(UpgradeStep):
            def __call__(self):
                self.update_workflow_security(
                    ['plone_workflow'], reindex_security=True)
        Step(self.portal_setup)

        self.assertEquals(['Anonymous'],
                          self.get_allowed_roles_and_users_for(folder))


    def test_update_workflow_security_expects_list_of_workflows(self):
        class Step(UpgradeStep):
            def __call__(self):
                self.update_workflow_security('foo')

        with self.assertRaises(ValueError) as cm:
            Step(self.portal_setup)

        self.assertEquals('"workflows" must be a list of workflow names.',
                          str(cm.exception))

    def set_workflow_chain(self, for_type, to_workflow):
        wftool = getToolByName(self.portal, 'portal_workflow')
        wftool.setChainForPortalTypes((for_type,),
                                      (to_workflow,))

    def assert_permission_acquired(self, permission, obj):
        not_acquired_permissions = self.get_not_acquired_permissions_of(obj)

        self.assertNotIn(
            permission, not_acquired_permissions,
            'Expected permission "%s" to be acquired on %s' % (
                permission, str(obj)))

    def assert_permission_not_acquired(self, permission, obj):
        not_acquired_permissions = self.get_not_acquired_permissions_of(obj)

        self.assertIn(
            permission, not_acquired_permissions,
            'Expected permission "%s" to NOT be acquired on %s' % (
                permission, str(obj)))

    def get_not_acquired_permissions_of(self, obj):
        acquired_permissions = filter(
            lambda item: not item.get('acquire'),
            obj.permission_settings())

        return map(lambda item: item.get('name'), acquired_permissions)

    def get_allowed_roles_and_users_for(self, obj):
        catalog = getToolByName(self.portal, 'portal_catalog')
        path = '/'.join(obj.getPhysicalPath())
        rid = catalog.getrid(path)
        index_data = catalog.getIndexDataForRID(rid)
        return index_data.get('allowedRolesAndUsers')

    def read_log(self):
        self.log.seek(0)
        return self.log.read().strip().split('\n')



from ftw.builder import Builder
from ftw.upgrade.interfaces import IPostUpgrade
from ftw.upgrade.tests.base import UpgradeTestCase
from zope.interface import Interface


class TestPostUpgrade(UpgradeTestCase):

    def test_post_upgrade_adapters_are_executed(self):
        self.package.with_profile(Builder('genericsetup profile')
                                  .with_upgrade(self.default_upgrade()))

        with self.package_created():
            self.install_profile('the.package:default', version='1000')

            execution_info = {}
            def post_upgrade_adapter(portal, request):
                execution_info.update({
                        'executed': True,
                        'portal_argument': portal,
                        'request_argument': request})

            self.portal.getSiteManager().registerAdapter(
                post_upgrade_adapter,
                required=(Interface, Interface),
                provided=IPostUpgrade,
                name='name-is-not-really-relevant')

            self.install_profile_upgrades('the.package:default')
            self.assertDictEqual(
                {'executed': True,
                 'portal_argument': self.portal,
                 'request_argument': self.portal.REQUEST},
                execution_info)

    def test_post_upgrade_adapters_are_executed_in_order_of_dependencies(self):
        self.package.with_profile(Builder('genericsetup profile')
                                  .named('addon'))
        self.package.with_profile(Builder('genericsetup profile')
                                  .named('product')
                                  .with_dependencies('the.package:addon'))
        self.package.with_profile(Builder('genericsetup profile')
                                  .named('customization')
                                  .with_dependencies('the.package:product')
                                  .with_upgrade(self.default_upgrade()))

        with self.package_created():
            self.install_profile('the.package:customization', version='1000')

            site_manager = self.portal.getSiteManager()
            execution_order = []

            def register_post_component_adapter(profile_id):
                class PostComponentAdapter(object):
                    def __init__(self, portal, request):
                        pass
                    def __call__(self):
                        execution_order.append(profile_id)
                site_manager.registerAdapter(PostComponentAdapter,
                                             required=(Interface, Interface),
                                             provided=IPostUpgrade,
                                             name=profile_id)

            map(register_post_component_adapter,
                ('the.package:addon', 'the.package:product', 'the.package:customization'))

            self.assertEquals([], execution_order)
            self.install_profile_upgrades('the.package:customization')

            self.assertEquals(['the.package:addon',
                               'the.package:product',
                               'the.package:customization'],
                              execution_order)



from ftw.upgrade.testing import INTID_UPGRADE_FUNCTIONAL_TESTING
from ftw.upgrade.tests.test_upgrade_step import TestUpgradeStep


class TestUpgradeStepIntids(TestUpgradeStep):

    layer = INTID_UPGRADE_FUNCTIONAL_TESTING



from datetime import datetime
from ftw.builder import Builder
from ftw.builder import create
from ftw.upgrade.tests.base import CommandAndInstanceTestCase
from ftw.upgrade.tests.helpers import no_logging_threads
from persistent.list import PersistentList
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
import os
import transaction


class TestInstallCommand(CommandAndInstanceTestCase):

    def setUp(self):
        super(TestInstallCommand, self).setUp()
        self.write_zconf_with_test_instance()

    def test_help(self):
        self.upgrade_script('install --help')

    def test_install_proposed_upgrades(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2011, 1, 1))
                          .named('The upgrade')))

        with self.package_created():
            self.install_profile('the.package:default', version='1')
            self.clear_recorded_upgrades('the.package:default')

            self.assertFalse(self.is_installed('the.package:default', datetime(2011, 1, 1)))
            exitcode, output = self.upgrade_script('install -s plone --proposed')
            self.assertEquals(0, exitcode)
            transaction.begin()  # sync transaction
            self.assertTrue(self.is_installed('the.package:default', datetime(2011, 1, 1)))
            self.assertIn('Result: SUCCESS', output)

    def test_install_failure_raises_exitcode(self):
        def failing_upgrade(setup_context):
            raise KeyError('foo')

        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('plone upgrade step')
                          .upgrading('1', to='2')
                          .calling(failing_upgrade)))

        with self.package_created():
            self.install_profile('the.package:default', version='1')
            with no_logging_threads():
                exitcode, output = self.upgrade_script('install -s plone --proposed',
                                                       assert_exitcode=False)
            self.assertEquals(3, exitcode)
            self.assertIn('Result: FAILURE', output)

    def test_install_list_of_upgrades(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2011, 1, 1))
                          .named('The upgrade')))

        with self.package_created():
            self.install_profile('the.package:default', version='1')
            self.clear_recorded_upgrades('the.package:default')

            self.assertFalse(self.is_installed('the.package:default', datetime(2011, 1, 1)))
            exitcode, output = self.upgrade_script(
                'install -s plone --upgrades 20110101000000@the.package:default')
            self.assertEquals(0, exitcode)
            transaction.begin()  # sync transaction
            self.assertTrue(self.is_installed('the.package:default', datetime(2011, 1, 1)))
            self.assertIn('Result: SUCCESS', output)

    def test_virtual_host_monster_is_configured_by_environment_variable(self):
        os.environ['UPGRADE_PUBLIC_URL'] = 'https://foo.bar.com/baz'
        self.layer['portal'].upgrade_info = PersistentList()

        setRoles(self.layer['portal'], TEST_USER_ID, ['Manager'])
        create(Builder('folder').with_id('the-folder'))

        def upgrade_step(setup_context):
            portal = setup_context.portal_url.getPortalObject()
            folder = portal.get('the-folder')
            portal.upgrade_info.append(folder.absolute_url())

        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('plone upgrade step').upgrading('1', to='2')
                          .calling(upgrade_step)))

        with self.package_created():
            self.install_profile('the.package:default', version='1')
            exitcode, output = self.upgrade_script('install -s plone --proposed')
            self.assertEquals(0, exitcode)
            transaction.begin()  # sync transaction
            self.assertEquals(['https://foo.bar.com/baz/the-folder'],
                              self.layer['portal'].upgrade_info)



from AccessControl.SecurityInfo import ClassSecurityInformation
from ftw.upgrade.interfaces import IExecutioner
from ftw.upgrade.interfaces import IPostUpgrade
from ftw.upgrade.interfaces import IUpgradeInformationGatherer
from ftw.upgrade.resource_registries import recook_resources
from ftw.upgrade.transactionnote import TransactionNote
from ftw.upgrade.utils import format_duration
from ftw.upgrade.utils import get_sorted_profile_ids
from Products.CMFCore.utils import getToolByName
from Products.GenericSetup.interfaces import ISetupTool
from Products.GenericSetup.upgrade import _upgrade_registry
from zope.component import adapts
from zope.component import getAdapters
from zope.interface import implements
import logging
import time
import transaction


logger = logging.getLogger('ftw.upgrade')


class Executioner(object):

    implements(IExecutioner)
    adapts(ISetupTool)
    security = ClassSecurityInformation()

    def __init__(self, portal_setup):
        self.portal_setup = portal_setup

    security.declarePrivate('install')
    def install(self, data):
        self._register_after_commit_hook()

        for profileid, upgradeids in data:
            self._upgrade_profile(profileid, upgradeids)

        for adapter in self._get_sorted_post_upgrade_adapters():
            adapter()

        TransactionNote().set_transaction_note()
        recook_resources()

    security.declarePrivate('install_upgrades_by_api_ids')
    def install_upgrades_by_api_ids(self, *upgrade_api_ids):
        gatherer = IUpgradeInformationGatherer(self.portal_setup)
        upgrades = gatherer.get_upgrades_by_api_ids(*upgrade_api_ids)
        data = [(upgrade['profile'], [upgrade['id']]) for upgrade in upgrades]
        return self.install(data)

    security.declarePrivate('install_profiles_by_profile_ids')
    def install_profiles_by_profile_ids(self, *profile_ids):
        for profile_id in profile_ids:
            # Starting from GenericSetup 1.8.0 getLastVersionForProfile can
            # handle profile ids with or without 'profile-' prefix, but we need
            # to support older versions as well, which only support it without
            # the prefix.
            prefix = 'profile-'
            if profile_id.startswith(prefix):
                profile_id = profile_id[len(prefix):]
            installed = self.portal_setup.getLastVersionForProfile(profile_id)
            if installed != 'unknown':
                logger.info('Ignoring already installed profile %s.',
                            profile_id)
                continue
            logger.info('Installing profile %s.', profile_id)
            # For runAllImportStepsFromProfile we still must have 'profile-' at
            # the start.
            self.portal_setup.runAllImportStepsFromProfile(prefix + profile_id)
            logger.info('Done installing profile %s.', profile_id)

    security.declarePrivate('_register_after_commit_hook')
    def _register_after_commit_hook(self):

        def notification_hook(success, *args, **kwargs):
            result = success and 'committed' or 'aborted'
            logger.info('Transaction has been %s.' % result)

        txn = transaction.get()
        txn.addAfterCommitHook(notification_hook)

    security.declarePrivate('_upgrade_profile')
    def _upgrade_profile(self, profileid, upgradeids):
        last_dest_version = None

        for upgradeid in upgradeids:
            last_dest_version = self._do_upgrade(profileid, upgradeid) \
                or last_dest_version

        self.portal_setup.setLastVersionForProfile(
            profileid, last_dest_version)

        self._set_quickinstaller_version(profileid)

    security.declarePrivate('_set_quickinstaller_version')
    def _set_quickinstaller_version(self, profileid):
        try:
            profileinfo = self.portal_setup.getProfileInfo(profileid)
        except KeyError:
            return

        quickinstaller = getToolByName(self.portal_setup,
                                       'portal_quickinstaller')
        product = profileinfo['product']
        if not quickinstaller.isProductInstalled(product):
            return

        version = quickinstaller.getProductVersion(product)
        if version:
            quickinstaller.get(product).installedversion = version

    security.declarePrivate('_do_upgrade')
    def _do_upgrade(self, profileid, upgradeid):
        start = time.time()

        step = _upgrade_registry.getUpgradeStep(profileid, upgradeid)
        logger.log(logging.INFO, '_' * 70)
        logger.log(logging.INFO, 'UPGRADE STEP %s: %s' % (
                profileid, step.title))

        step.doStep(self.portal_setup)
        TransactionNote().add_upgrade(profileid, step.dest, step.title)

        msg = "Ran upgrade step %s for profile %s" % (
            step.title, profileid)
        logger.log(logging.INFO, msg)

        logger.log(logging.INFO, 'Upgrade step duration: %s' % format_duration(
                time.time() - start))

        return step.dest

    security.declarePrivate('_get_sorted_post_upgrade_adapters')
    def _get_sorted_post_upgrade_adapters(self):
        """Returns a list of post upgrade adapters, sorted by
        profile dependencies.
        Assumes that the names of the adapters are profile names
        (e.g. "ftw.upgrade:default").
        """

        profile_order = get_sorted_profile_ids(self.portal_setup)

        portal_url = getToolByName(self.portal_setup, 'portal_url')
        portal = portal_url.getPortalObject()
        adapters = list(getAdapters((portal, portal.REQUEST), IPostUpgrade))

        def _sorter(a, b):
            name_a = a[0]
            name_b = b[0]

            if name_a not in profile_order and name_b not in profile_order:
                return 0

            elif name_a not in profile_order:
                return -1

            elif name_b not in profile_order:
                return 1

            else:
                return cmp(profile_order.index(name_a),
                           profile_order.index(name_b))

        adapters.sort(_sorter)
        return [adapter for name, adapter in adapters]



from ftw.upgrade.command.jsonapi import add_requestor_authentication_argument
from ftw.upgrade.command.jsonapi import add_site_path_argument
from ftw.upgrade.command.jsonapi import error_handling
from ftw.upgrade.command.jsonapi import with_api_requestor
from ftw.upgrade.command.terminal import TERMINAL


DOCS = """
{t.bold}DESCRIPTION:{t.normal}
    Recook CSS and JavaScript resource bundles.
[quote]
    $ ./bin/upgrade recook --site Plone
[/quote]
""".format(t=TERMINAL).strip()


def setup_argparser(commands):
    command = commands.add_parser(
        'recook',
        help='Recook CSS and JavaScript resource bundles.',
        description=DOCS)
    command.set_defaults(func=recook_command)
    add_requestor_authentication_argument(command)
    add_site_path_argument(command)


@with_api_requestor
@error_handling
def recook_command(args, requestor):
    print requestor.POST('recook_resources').json()



from Products.CMFCore.utils import getToolByName
from zope.component.hooks import getSite


def recook_resources():
    for name in ('portal_css', 'portal_javascripts'):
        registry = getToolByName(getSite(), name)
        registry.cookResources()



from ftw.builder import Builder
from ftw.upgrade.executioner import Executioner
from ftw.upgrade.interfaces import IExecutioner
from ftw.upgrade.tests.base import UpgradeTestCase
from Products.CMFCore.utils import getToolByName
from zope.component import queryAdapter
from zope.interface.verify import verifyClass
import transaction


class TestExecutioner(UpgradeTestCase):

    def test_component_is_registered(self):
        setup_tool = getToolByName(self.layer['portal'], 'portal_setup')
        executioner = queryAdapter(setup_tool, IExecutioner)
        self.assertNotEqual(None, executioner)

    def test_implements_interface(self):
        verifyClass(IExecutioner, Executioner)

    def test_installs_upgrades(self):
        def upgrade(setup_context):
            portal = setup_context.portal_url.getPortalObject()
            portal.upgrade_step_executed = True

        self.package.with_profile(Builder('genericsetup profile')
                                   .with_upgrade(Builder('plone upgrade step')
                                                 .upgrading('1000', to='1002')
                                                 .calling(upgrade)))

        self.portal.upgrade_step_executed = False
        with self.package_created():
            self.install_profile('the.package:default', version='1000')
            self.assertFalse(self.portal.upgrade_step_executed)
            self.install_profile_upgrades('the.package:default')
            self.assertTrue(self.portal.upgrade_step_executed)

    def test_install_upgrades_by_api_ids(self):
        def upgrade(setup_context):
            portal = setup_context.portal_url.getPortalObject()
            portal.upgrade_step_executed = True

        self.package.with_profile(Builder('genericsetup profile')
                                   .with_upgrade(Builder('plone upgrade step')
                                                 .upgrading('1000', to='1002')
                                                 .calling(upgrade)))

        self.portal.upgrade_step_executed = False
        with self.package_created():
            self.install_profile('the.package:default', version='1000')
            self.assertFalse(self.portal.upgrade_step_executed)
            executioner = queryAdapter(self.portal_setup, IExecutioner)
            executioner.install_upgrades_by_api_ids('1002@the.package:default')
            self.assertTrue(self.portal.upgrade_step_executed)

    def test_transaction_note(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('plone upgrade step')
                          .upgrading('1000', to='1001')
                          .titled('Register "foo" utility'))
            .with_upgrade(Builder('plone upgrade step')
                          .upgrading('1001', to='1002')
                          .titled('Update email address'))
            .with_upgrade(Builder('plone upgrade step')
                          .upgrading('1002', to='1003')
                          .titled('Update email from name')))

        with self.package_created():
            self.install_profile('the.package:default', version='1000')
            self.install_profile_upgrades('the.package:default')
            self.assertMultiLineEqual(
                u'the.package:default -> 1001 (Register "foo" utility)\n'
                u'the.package:default -> 1002 (Update email address)\n'
                u'the.package:default -> 1003 (Update email from name)',
                transaction.get().description)

    def test_after_commit_hook(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('plone upgrade step')
                          .upgrading('1000', to='1001')
                          .titled('Register "foo" utility')))

        with self.package_created():
            self.install_profile('the.package:default', version='1000')
            self.install_profile_upgrades('the.package:default')

            hooks = list(transaction.get().getAfterCommitHooks())
            hook_funcs = [h[0] for h in hooks]

            self.assertIn(
                'notification_hook',
                [f.func_name for f in hook_funcs],
                'Our notification_hook should be registered')

            transaction.commit()
            self.assertEquals(
                [],
                list(transaction.get().getAfterCommitHooks()),
                'Hook registrations should not persist across transactions')

    def test_resources_are_recooked_after_installing_upgrades(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('plone upgrade step').upgrading('1000', to='1001')))
        with self.package_created():
            self.install_profile('the.package:default', version='1000')
            with self.assert_resources_recooked():
                self.install_profile_upgrades('the.package:default')

    def test_updates_quickinstaller_version(self):
        quickinstaller = getToolByName(self.portal, 'portal_quickinstaller')

        self.package.with_profile(Builder('genericsetup profile')
                                  .with_upgrade(Builder('plone upgrade step')
                                                .upgrading('1000', to='1001')))
        self.package.with_version('1.1')

        with self.package_created():
            self.install_profile('the.package:default', version='1000')
            quickinstaller.get('the.package').installedversion = '1.0'
            self.assertEquals('1.0', quickinstaller.get('the.package').getInstalledVersion())
            self.install_profile_upgrades('the.package:default')
            self.assertEquals('1.1', quickinstaller.get('the.package').getInstalledVersion())



from datetime import datetime
from ftw.builder import Builder
from ftw.builder import create
from ftw.upgrade.tests.base import UpgradeTestCase
from Products.CMFPlone.interfaces import IMigratingPloneSiteRoot
from Products.GenericSetup.interfaces import EXTENSION
from zope.configuration.config import ConfigurationExecutionError


class TestDirectoryMetaDirective(UpgradeTestCase):

    def setUp(self):
        super(TestDirectoryMetaDirective, self).setUp()
        self.profile = Builder('genericsetup profile')
        self.package.with_profile(self.profile)

    def test_upgrade_steps_are_registered(self):
        self.profile.with_upgrade(Builder('ftw upgrade step')
                                  .to(datetime(2011, 1, 1, 8))
                                  .named('add_action'))
        self.profile.with_upgrade(Builder('ftw upgrade step')
                                  .to(datetime(2011, 2, 2, 8))
                                  .named('remove_action'))

        with self.package_created():
            self.assert_upgrades([
                    {'source': ('10000000000000',),
                     'dest': ('20110101080000',),
                     'title': u'Add action.'},

                    {'source': ('20110101080000',),
                     'dest': ('20110202080000',),
                     'title': u'Remove action.'}])

    def test_first_source_version_is_last_regulare_upgrade_step(self):
        self.profile.with_upgrade(Builder('plone upgrade step')
                                  .upgrading('1000', to='1001')
                                  .titled('Register foo utility.'))
        self.profile.with_upgrade(Builder('ftw upgrade step')
                                  .to(datetime(2011, 1, 1, 8))
                                  .named('add_action'))

        with self.package_created():
            self.assert_upgrades([
                    {'source': ('1000',),
                     'dest': ('1001',),
                     'title': u'Register foo utility.'},

                    {'source': ('1001',),
                     'dest': ('20110101080000',),
                     'title': u'Add action.'}])

    def test_registers_migration_generic_setup_profile_foreach_step(self):
        self.profile.with_upgrade(Builder('ftw upgrade step')
                                  .to(datetime(2011, 1, 1, 8))
                                  .named('add_an_action'))

        with self.package_created() as package:
            upgrade_path = package.package_path.joinpath(
                'upgrades', '20110101080000_add_an_action')
            self.assert_profile(
                {'id': 'the.package.upgrades:default-upgrade-20110101080000',
                 'title': 'Upgrade the.package:default ' + \
                     'to 20110101080000: Add an action.',
                 'description': '',
                 'path': upgrade_path,
                 'product': 'the.package.upgrades',
                 'type': EXTENSION,
                 'for': IMigratingPloneSiteRoot})

    def test_package_modules_is_not_corrupted(self):
        # Regression: when the upgrade-step:directory directive is used from
        # the package-directory with a relative path (directory="upgrades"),
        # it corrupted the sys.modules entry of the package.

        package_builder = (
            Builder('python package')
            .named('other.package')
            .at_path(self.layer['temp_directory'])
            .with_file('__init__.py', 'PACKAGE = "package root"')

            .with_directory('profiles/default')
            .with_zcml_node('genericsetup:registerProfile',
                            name='default',
                            title='other.package:default',
                            directory='profiles/default',
                            provides='Products.GenericSetup.interfaces.EXTENSION')

            .with_directory('upgrades')
            .with_file('upgrades/__init__.py', 'PACKAGE = "upgrades package"')
            .with_zcml_include('ftw.upgrade', file='meta.zcml')
            .with_zcml_node('upgrade-step:directory',
                            profile='other.package:default',
                            directory='upgrades'))

        with create(package_builder).zcml_loaded(self.layer['configurationContext']):
            import other.package
            self.assertEquals('package root', other.package.PACKAGE)

    def test_profile_must_be_registed_before_registering_upgrade_directory(self):
        package_builder = (Builder('python package')
                           .named('other.package')
                           .at_path(self.layer['temp_directory'])
                           .with_zcml_include('ftw.upgrade', file='meta.zcml')
                           .with_zcml_node('upgrade-step:directory',
                                           profile='other.package:default',
                                           directory='.'))

        with create(package_builder) as package:
            with self.assertRaises(ConfigurationExecutionError) as cm:
                package.load_zcml(self.layer['configurationContext'])

        self.assertEqual(
            "<class 'ftw.upgrade.exceptions.UpgradeStepConfigurationError'>: "
            'The profile "other.package:default" needs to be registered'
            ' before registering its upgrade step directory.',
            str(cm.exception).splitlines()[0])

    def test_profile_version_is_set_to_latest_profile_version(self):
        self.profile.with_upgrade(Builder('ftw upgrade step').to(datetime(2011, 1, 1, 8)))
        self.profile.with_upgrade(Builder('ftw upgrade step').to(datetime(2011, 2, 2, 8)))

        with self.package_created() as package:
            profile_path = package.package_path.joinpath('profiles', 'default')
            self.assert_profile(
                {'id': u'the.package:default',
                 'title': u'the.package',
                 'description': u'',
                 'path': profile_path,
                 'version': '20110202080000',
                 'product': 'the.package',
                 'type': EXTENSION,
                 'for': None})

    def test_version_set_to_default_when_no_upgrades_defined(self):
        upgrades = self.package.package.get_subpackage('upgrades')
        upgrades.with_zcml_include('ftw.upgrade', file='meta.zcml')
        upgrades.with_zcml_node('upgrade-step:directory',
                                profile='the.package:default',
                                directory='.')

        with self.package_created() as package:
            profile_path = package.package_path.joinpath('profiles', 'default')
            self.assert_profile(
                {'id': u'the.package:default',
                 'title': u'the.package',
                 'description': u'',
                 'path': profile_path,
                 'version': u'10000000000000',
                 'product': 'the.package',
                 'type': EXTENSION,
                 'for': None})

    def test_profile_must_not_have_a_metadata_version_defined(self):
        self.profile.with_fs_version('1000')
        self.profile.with_upgrade(Builder('ftw upgrade step').to(datetime(2011, 1, 1, 8)))

        with create(self.package) as package:
            with self.assertRaises(ConfigurationExecutionError) as cm:
                package.load_zcml(self.layer['configurationContext'])

        self.assertEqual(
            "<class 'ftw.upgrade.exceptions.UpgradeStepConfigurationError'>: "
            'Registering an upgrades directory for "the.package:default" requires'
            ' this profile to not define a version in its metadata.xml.'
            ' The version is automatically set to the latest upgrade.',
            str(cm.exception).splitlines()[0])

    def assert_upgrades(self, expected):
        upgrades = self.portal_setup.listUpgrades('the.package:default')
        got = [dict((key, value) for (key, value) in step.items()
                    if key in ('source', 'dest', 'title'))
               for step in upgrades]
        self.maxDiff = None
        self.assertItemsEqual(expected, got)

    def assert_profile(self, expected):
        self.assertTrue(
            self.portal_setup.profileExists(expected['id']),
            'Profile "{0}" does not exist. Profiles: {1}'.format(
                expected['id'],
                [profile['id'] for profile in self.portal_setup.listProfileInfo()]))

        got = self.portal_setup.getProfileInfo(expected['id'])
        self.maxDiff = None
        self.assertDictEqual(expected, got)



from ftw.upgrade.command import create
from ftw.upgrade.command import help
from ftw.upgrade.command import install
from ftw.upgrade.command import list_cmd
from ftw.upgrade.command import plone_upgrade
from ftw.upgrade.command import recook
from ftw.upgrade.command import sites
from ftw.upgrade.command import touch
from ftw.upgrade.command import user
from ftw.upgrade.command.formatter import FlexiFormatter
from ftw.upgrade.command.terminal import TERMINAL
from ftw.upgrade.command.utils import capture
from pkg_resources import get_distribution
import argcomplete
import argparse
import logging
import json
import sys


VERSION = get_distribution('ftw.upgrade').version
logger = logging.getLogger('ftw.upgrade')


DOCS = """
{t.bold}DESCRIPTION:{t.normal}
    The bin/upgrade script helps to create upgrade steps as well as to \
inspect the upgrades of a Plone site and install upgrades. \
The command is installed by the "ftw.upgrade" Plone addon package.

{t.bold}CREATING AND MANAGING UPGRADES:{t.normal}
    For using the "create" and "touch" command to manage upgrade steps on the \
file system, the package under development must use the \
"upgrade-step:directory" ZCML directive to declare an upgrade step directory, \
which should be named "upgrades".

    The "create" command then allows to easily generate upgrades with the \
current timestamp as destination version, the "touch" command on the other \
hand allows to change the order of upgrades by updating the timestamp \
(destination version).

{t.bold}LISTING AND INSTALLING UPGRADES:{t.normal}
    In order to use the commands "sites", "list" and "install" a Zope \
instance must be running, where "ftw.upgrade" is available and the \
"upgrades-api" endpoint is available.

{t.bold}ZOPE INSTANCE DISCOVERY:{t.normal}
    The Zope instance is discovered automatically by searching for all \
"parts/*/etc/zope.conf" files in the buildout directory, looking up \
the instance port and testing whether the port is bound on localhost.
If multiple instances are running, the first one running is used.

{t.bold}AUTHENTICATION:{t.normal}
    There are three authentication options which are tested and used in this \
order:

    - "--auth" argument for basic authentication
    - "UPGRADE_AUTHENTICATION" environment variable for basic authentication
    - automatic tempfile based authorization negotiation, using the \
      automatically generated manager user "system-upgrade".

    {t.bold}"--auth" argument authentication{t.normal}
    Commands requiring authentication have a "--auth" argument which precede \
the tempfile negotiation mechanism. \
It uses basic authentication. \
The value is expected to be "<username>:<password>". \
The user should have the "Manager" role.

    {t.bold}"UPGRADE_AUTHENTICATION" environment variable{t.normal}
    When no "--auth" argument is used, the "UPGRADE_AUTHENTICATION" \
environment variable is tested for authentication information. \
This works exactly like the "--auth" argument.

    {t.bold}Tempfile negotiation authentication{t.normal}
    The automatic tempfile negotiation is used when neither the "--auth" \
argument nor the "UPGRADE_AUTHENTICATION" environment variable is used. \
It creates a tempfile, readable only for the owner user and writes a random \
hash to the file. The filename and the hash is submitted as request header \
and the backend reads the tempfile and verifies the content.
    The idea is that if a user has access to the machine with the user \
running Zope, he can easily create an emergency administator user from the \
command line and acquire full manager access to the Zope installation. \
By being able to write the tempfile as this user on the server machine \
the access is verified.
    When using this mechanism, a manager user is created at Zope app level, \
named "system-upgrade". The user has a random password. All requests are \
then authenticated with this user.
    The temporary file is written to \
[buildout-directory]var/ftw.upgrade-authentication

{t.bold}VIRTUAL HOSTING:{t.normal}
    For some upgrade steps it is important that "absolute_url()" returns a \
public URL, for example when contacting external services. \
By setting the environment variable "UPGRADE_PUBLIC_URL", "bin/upgrade" \
automatically configures the virtual host monster. \
Be aware that the Plone site URL part will not be part of the public URL. \
This means the public URL must point to the Plone site which is selected \
with the "--site" argument.

    Examples:
[quote]
$ UPGRADE_PUBLIC_URL="https://my.site.com/" bin/upgrade
$ UPGRADE_PUBLIC_URL="http://my.site.com/foo/bar" bin/upgrade
[/quote]

{t.bold}MORE INFORMATION:{t.normal}
    Project Homepage: https://github.com/4teamwork/ftw.upgrade
    ftw.upgrade version: {version}
""".format(t=TERMINAL, version=VERSION).strip()


class UpgradeArgumentParser(argparse.ArgumentParser):
    def __init__(self, *args, **kwargs):
        kwargs['formatter_class'] = FlexiFormatter
        super(UpgradeArgumentParser, self).__init__(*args, **kwargs)


class UpgradeCommand(object):

    def __init__(self):
        self.parser = UpgradeArgumentParser(
            sys.argv[0],
            epilog=DOCS)

        self.parser.add_argument('--version', action='version',
                                 version='%(prog)s {0}'.format(VERSION))

        argcomplete.autocomplete(self.parser)

        commands = self.parser.add_subparsers(help='Command')
        create.setup_argparser(commands)
        install.setup_argparser(commands)
        list_cmd.setup_argparser(commands)
        plone_upgrade.setup_argparser(commands)
        recook.setup_argparser(commands)
        sites.setup_argparser(commands)
        touch.setup_argparser(commands)
        user.setup_argparser(commands)

        # Register as last.
        help.setup_argparser(commands)

    def __call__(self):
        args = self.parser.parse_args()
        configure_logging(args)
        setattr(args, 'parser', self.parser)
        if getattr(args, 'all_sites', False):
            info = {}
            while True:
                try:
                    with capture() as out:
                        args.func(args)
                    output = out.getvalue().strip()
                    if getattr(args, 'json', False):
                        # "[]" must be turned into [] and put in dictionary.
                        output = json.loads(output)
                        info[args.picked_site] = output
                    else:
                        logger.info('Acting on site {0}'.format(
                            args.picked_site))
                        print(output)
                except StopIteration:
                    break
            if getattr(args, 'json', False):
                # Pretty print the output of all sites.
                print(json.dumps(info, indent=4, encoding='utf-8'))
        else:
            args.func(args)


def configure_logging(args):
    # Extend the level names with colors.
    logging.addLevelName(
        logging.INFO, '{t.green}{levelname}{t.normal}'.format(
            t=TERMINAL, levelname=logging.getLevelName(logging.INFO)))
    for level in (logging.WARN, logging.ERROR, logging.CRITICAL):
        logging.addLevelName(
            level, '{t.red_bold}{levelname}{t.normal}'.format(
                t=TERMINAL, levelname=logging.getLevelName(level)))
    if getattr(args, 'verbose', False):
        # Log debug level and higher for all loggers.
        start_level = logging.DEBUG
        format = '%(levelname)s %(name)s: %(message)s'
        logging.basicConfig(level=start_level, format=format)
    else:
        # Log info level and higher, only for our own logger.
        start_level = logging.INFO
        format = '%(levelname)s: %(message)s'
        logging.basicConfig(level=start_level, format=format)
        for handler in logging.root.handlers:
            handler.addFilter(logging.Filter('ftw.upgrade'))


def main():
    UpgradeCommand()()



from datetime import datetime
from path import Path
import inflection
import os


PYTHON_TEMPLATE = '''from ftw.upgrade import UpgradeStep


class {classname}(UpgradeStep):
    """{docstring}.
    """

    def __call__(self):
        self.install_upgrade_profile()
'''

DATETIME_FORMAT = '%Y%m%d%H%M%S'


class UpgradeStepCreator(object):

    def __init__(self, upgrades_directory):
        self.upgrades_directory = upgrades_directory

    def create(self, name):
        name = name.replace(' ', '_').replace('\.$', '')

        step_name = '{0}_{1}'.format(
            datetime.now().strftime(DATETIME_FORMAT),
            inflection.underscore(name))
        step_directory = os.path.join(self.upgrades_directory, step_name)
        os.mkdir(step_directory)

        Path(step_directory).joinpath('__init__.py').touch()

        code_path = os.path.join(step_directory, 'upgrade.py')
        with open(code_path, 'w+') as code_file:
            code_file.write(
                PYTHON_TEMPLATE.format(
                    classname=inflection.camelize(name),
                    docstring=inflection.humanize(
                        inflection.underscore(name))))

        return step_directory



from datetime import datetime
from ftw.testing import freeze
from ftw.upgrade.directory.scaffold import UpgradeStepCreator
from unittest2 import TestCase
import os.path
import re
import shutil
import tempfile


class TestUpgradeStepCreator(TestCase):

    def setUp(self):
        self.upgrades_directory = tempfile.mkdtemp('ftw.upgrade.tests')

    def tearDown(self):
        shutil.rmtree(self.upgrades_directory)

    def test_generates_directory_and_upgrade_code(self):
        with freeze(datetime(2014, 11, 14, 22, 44, 55)):
            UpgradeStepCreator(self.upgrades_directory).create('AddControlpanelAction')

        upgrade_directory = os.path.join(self.upgrades_directory,
                                         '20141114224455_add_controlpanel_action')
        upgrade_code_file = os.path.join(upgrade_directory, 'upgrade.py')

        self.assertTrue(os.path.isdir(upgrade_directory))
        self.assertTrue(os.path.isfile(upgrade_code_file))
        with open(upgrade_code_file) as python_file:
            code = python_file.read()

        self.maxDiff = None
        self.assertMultiLineEqual(
            '\n'.join((
                    'from ftw.upgrade import UpgradeStep',
                    '',
                    '',
                    'class AddControlpanelAction(UpgradeStep):',
                    '    """Add controlpanel action.',
                    '    """',
                    '',
                    '    def __call__(self):',
                    '        self.install_upgrade_profile()',
                    '')),
            code)

    def test_generate_upgrade_step_by_camelcase_name(self):
        with freeze(datetime(2014, 11, 14, 22, 44, 55)):
            UpgradeStepCreator(self.upgrades_directory).create('AddControlpanelAction')

        self.assert_upgrade({'name': '20141114224455_add_controlpanel_action',
                             'classname': 'AddControlpanelAction',
                             'docstring': 'Add controlpanel action.'})

    def test_generate_upgrade_step_by_underscore_name(self):
        with freeze(datetime(2014, 11, 14, 22, 44, 55)):
            UpgradeStepCreator(self.upgrades_directory).create('add_controlpanel_action')

        self.assert_upgrade({'name': '20141114224455_add_controlpanel_action',
                             'classname': 'AddControlpanelAction',
                             'docstring': 'Add controlpanel action.'})

    def test_generate_upgrade_step_by_humanized_name(self):
        with freeze(datetime(2014, 11, 14, 22, 44, 55)):
            UpgradeStepCreator(self.upgrades_directory).create('Add controlpanel action')

        self.assert_upgrade({'name': '20141114224455_add_controlpanel_action',
                             'classname': 'AddControlpanelAction',
                             'docstring': 'Add controlpanel action.'})

    def assert_upgrade(self, expected):
        upgrade_directory = os.path.join(self.upgrades_directory, expected['name'])
        upgrade_code_file = os.path.join(upgrade_directory, 'upgrade.py')

        self.assertTrue(os.path.isdir(upgrade_directory))
        self.assertTrue(os.path.isfile(upgrade_code_file))

        with open(upgrade_code_file) as python_file:
            code = python_file.read()

        got = {'name': expected['name'],
               'classname': re.search('class ([^\(]*)\(UpgradeStep\):', code).group(1),
               'docstring': re.compile('"""(.*?)"""', re.DOTALL).search(code).group(1).strip()}
        self.assertDictEqual(expected, got)





class CyclicDependencies(Exception):
    """Cyclic dependencies detected.
    """

    def __init__(self, dependencies, cyclic_dependencies=()):
        super(CyclicDependencies, self).__init__(
            'Cyclic dependencies detected.')
        self.dependencies = dependencies
        self.cyclic_dependencies = cyclic_dependencies


class UpgradeNotFound(Exception):
    """An upgrade could not be found.
    """

    def __init__(self, api_id):
        self.api_id = api_id
        super(UpgradeNotFound, self).__init__(
            'The upgrade "{0}" could not be found.'.format(api_id))


class UpgradeStepDefinitionError(Exception):
    """An upgrade step definition is wrong.
    """


class UpgradeStepConfigurationError(Exception):
    """The upgrade steps directory configuration is wrong.
    """


class NoAssociatedProfileError(ValueError):
    """The upgrade step has no associated profile.
    """

    def __init__(self):
        super(NoAssociatedProfileError, self).__init__(
            self, NoAssociatedProfileError.__doc__.strip())



from ftw.upgrade.command.terminal import TERMINAL
import argparse
import os
import pydoc
import sys


DOCS = """
{t.bold}DESCRIPTION:{t.normal}
    Display help information about the bin/upgrade commands.

{t.bold}EXAMPLES:{t.normal}
[quote]
$ bin/upgrade help
$ bin/upgrade help create
[/quote]
""".format(t=TERMINAL).strip()


def setup_argparser(commands):
    command = commands.add_parser('help',
                                  help='Display help information.',
                                  description=DOCS)
    command.set_defaults(func=help_command)

    command.add_argument('command', nargs='?',
                         choices=get_commands(commands.container).keys(),
                         help='Command to describe.')


def help_command(args):
    commands = get_commands(args.parser)
    parser = commands.get(args.command, args.parser)
    if sys.stdout.isatty():
        if os.system('(less -R) 2>/dev/null') == 0:
            return pydoc.pipepager(parser.format_help(), cmd='less -R')
    else:
        print parser.format_help()


def get_commands(parser):
    commands = {}
    for action in parser._actions:
        if not isinstance(action, argparse._SubParsersAction):
            continue
        commands.update(action.choices)
    return commands



from ftw.builder import Builder
from ftw.builder import create
from ftw.upgrade.tests.base import CommandTestCase


class TestCreateCommand(CommandTestCase):

    def test_creating_an_upgrade_step(self):
        package = create(Builder('python package')
                         .named('the.package')
                         .at_path(self.layer.sample_buildout)
                         .with_directory('upgrades'))

        self.upgrade_script('create AddControlpanelAction')

        upgrades_dir = package.package_path.joinpath('upgrades')
        self.assertEqual(
            1, len(upgrades_dir.listdir()),
            'Expected exactly one directory to be generated, got {0}'.format(
                upgrades_dir.listdir()))

        step_path, = upgrades_dir.listdir()
        self.assertRegexpMatches(step_path.name, r'^\d{14}_add_controlpanel_action$')

        code_path = step_path.joinpath('upgrade.py')
        self.assertTrue(code_path.exists(), 'upgrade.py is missing')
        self.assertTrue(step_path.joinpath('__init__.py').exists(),
                        'There is no __init__.py in the upgrade directory.'
                        ' It is important so that plone.reload works.')
        self.maxDiff = True
        self.assertIn('class AddControlpanelAction(UpgradeStep):', code_path.text())

    def test_creating_an_upgrade_step_with_specifying_upgrades_directory(self):
        package = create(Builder('python package')
                         .named('the.package')
                         .at_path(self.layer.sample_buildout)
                         .with_directory('upgrades')
                         .with_directory('subpackage/upgrades'))

        upgrades_dir = package.package_path.joinpath('upgrades')
        subpackage_upgrades_dir = package.package_path.joinpath('subpackage', 'upgrades')
        self.upgrade_script('create AddControlpanelAction --path {0}'.format(
                subpackage_upgrades_dir))

        self.assertEqual(
            0, len(upgrades_dir.listdir()),
            'Expected default upgrades directory to be empty; {0}'.format(
                upgrades_dir.listdir()))

        self.assertEqual(
            1, len(subpackage_upgrades_dir.listdir()),
            'Expected subpackage upgrades directory to have one upgrade; {0}'.format(
                subpackage_upgrades_dir.listdir()))

    def test_fails_when_no_egginfo_found(self):
        exitcode, output = self.upgrade_script('create Title', assert_exitcode=False)
        self.assertEqual(1, exitcode, 'command should fail because there is no egg-info')
        self.assertIn('WARNING: no *.egg-info directory could be found.',
                      output)
        self.assertIn('ERROR: Please provide the path to the upgrades directory with --path.',
                      output)

    def test_fails_when_no_upgrades_directory_found(self):
        create(Builder('python package')
               .named('the.package')
               .at_path(self.layer.sample_buildout))

        exitcode, output = self.upgrade_script('create Title', assert_exitcode=False)
        self.assertEqual(1, exitcode, 'command should fail because there is no upgrades directory')
        self.assertIn('WARNING: no "upgrades" directory could be found.',
                      output)
        self.assertIn('ERROR: Please provide the path to the upgrades directory with --path.',
                      output)



from BTrees.OOBTree import OOBTree
from ftw.upgrade.interfaces import IUpgradeStepRecorder
from Products.CMFPlone.interfaces import IPloneSiteRoot
from zope.annotation import IAnnotations
from zope.component import adapts
from zope.interface import implements
from zope.interface import Interface


ANNOTATION_KEY = 'ftw.upgrade:recorder'


class UpgradeStepRecorder(object):
    implements(IUpgradeStepRecorder)
    adapts(IPloneSiteRoot, Interface)

    def __init__(self, portal, profilename):
        self.portal = portal
        self.profile = self._normalize_profilename(profilename)
        self.storage = self._get_profile_storage()

    def is_installed(self, target_version):
        return bool(self.storage.get(target_version, False))

    def mark_as_installed(self, target_version):
        self.storage[target_version] = True

    def _get_profile_storage(self):
        annotations = IAnnotations(self.portal)
        if ANNOTATION_KEY not in annotations:
            annotations[ANNOTATION_KEY] = OOBTree()

        if self.profile not in annotations[ANNOTATION_KEY]:
            annotations[ANNOTATION_KEY][self.profile] = OOBTree()

        return annotations[ANNOTATION_KEY][self.profile]

    def _normalize_profilename(self, profilename):
        if profilename.startswith('profile-'):
            profilename = profilename[len('profile-'):]

        if not isinstance(profilename, unicode):
            profilename = profilename.decode('utf-8')

        return profilename



from ftw.upgrade.interfaces import IRecordableHandler
from ftw.upgrade.interfaces import IUpgradeStepRecorder
from Products.CMFCore.utils import getToolByName
from zope.component import getMultiAdapter
from zope.interface import alsoProvides


def wrap_upgrade_step(handler, upgrade_profile, base_profile, target_version):
    def upgrade_step_wrapper(portal_setup):
        result = handler(portal_setup, upgrade_profile)

        portal = getToolByName(portal_setup, 'portal_url').getPortalObject()
        recorder = getMultiAdapter((portal, base_profile),
                                   IUpgradeStepRecorder)
        recorder.mark_as_installed(target_version)
        return result
    alsoProvides(upgrade_step_wrapper, IRecordableHandler)
    return upgrade_step_wrapper



from setuptools import setup, find_packages
import os

version = '1.16.2'

tests_require = [
    'unittest2',
    'mocker',
    'ftw.testing >= 1.8.1',
    'ftw.testbrowser',
    'ftw.builder >= 1.6.0',
    'plone.testing',
    'plone.app.testing',
    'plone.app.intid',

    'zope.configuration',
    'zc.recipe.egg',
    'transaction',
    'Products.ATContentTypes',
    'Products.CMFPlacefulWorkflow',
    ]

extras_require = {
    'colors': ['blessed <= 1.9.5'],
    'tests': tests_require}

setup(name='ftw.upgrade',
      version=version,
      description='An upgrade control panel and upgrade '
      'helpers for plone upgrades.',

      long_description=open('README.rst').read() + '\n' +
      open(os.path.join('docs', 'HISTORY.txt')).read(),

      classifiers=[
        'Framework :: Plone',
        'Framework :: Plone :: 4.1',
        'Framework :: Plone :: 4.2',
        'Framework :: Plone :: 4.3',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Programming Language :: Python',
        'Topic :: Software Development',
        ],

      keywords='plone ftw upgrade',
      author='4teamwork AG',
      author_email='mailto:info@4teamwork.ch',
      url='https://github.com/4teamwork/ftw.upgrade',
      license='GPL2',

      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['ftw'],
      include_package_data=True,
      zip_safe=False,

      install_requires=[
        'argcomplete',
        'argparse',
        'inflection',
        'path.py >= 6.2',
        'requests',
        'setuptools',
        'tarjan',

        # Zope
        'AccessControl',
        'Acquisition',
        'transaction',
        'Products.BTreeFolder2',
        'Products.ZCatalog',
        'zope.component',
        'zope.interface',
        'zope.publisher',
        'Zope2',

        # Plone
        'Plone',
        'Products.GenericSetup',
        'plone.browserlayer',
        'Products.CMFCore',
        'Products.CMFPlone',
        ],

      tests_require=tests_require,
      extras_require=extras_require,

      entry_points={
        'z3c.autoinclude.plugin': [
            'target = plone'],

        'console_scripts': [
            'upgrade = ftw.upgrade.command:main']
        })



from ftw.testing import MockTestCase
from ftw.upgrade.exceptions import CyclicDependencies
from ftw.upgrade.utils import find_cyclic_dependencies
from ftw.upgrade.utils import format_duration
from ftw.upgrade.utils import get_sorted_profile_ids
from ftw.upgrade.utils import SizedGenerator
from ftw.upgrade.utils import subject_from_docstring
from ftw.upgrade.utils import topological_sort
from unittest2 import TestCase


class TestTopologicalSort(TestCase):

    def test_simple(self):
        items = ['b', 'a', 'c']
        dependencies = (
            ('a', 'b'),
            ('b', 'c'))

        self.assertEqual(['a', 'b', 'c'],
                         topological_sort(items, dependencies))

    def test_advanced(self):
        items = ['a', 'c', 'b', 'd']
        dependencies = (
            ('a', 'b'),
            ('a', 'c'),
            ('b', 'd'),
            ('b', 'c'))

        self.assertEqual(['a', 'b', 'c', 'd'],
                         topological_sort(items, dependencies))

    def test_duplicated(self):
        items = ['a', 'b', 'a']
        dependencies = (
            ('b', 'a'),
            )

        self.assertEqual(['b', 'a'],
                         topological_sort(items, dependencies))

    def test_cyclic(self):
        items = ['a', 'b']
        dependencies = (
            ('a', 'b'),
            ('b', 'a'))

        self.assertEqual(None,
                         topological_sort(items, dependencies))

    def test_root_nodes_are_reversed_ordered(self):
        items = ['policy', 'xy', 'foo']

        self.assertEqual(
            ['xy', 'policy', 'foo'],
            topological_sort(items, (('policy', 'foo'),)))

        self.assertEqual(
            ['xy', 'policy', 'foo'],
            topological_sort(reversed(items), (('policy', 'foo'),)),
            'items input order should not change result order')


class TestFindCyclicDependencies(TestCase):

    def test_direct_cyclic_dependencies(self):
        dependencies = (
            ('bar', 'foo'),
            ('baz', 'foo'),
            ('baz', 'bar'),
            ('foo', 'bar'),
            ('default', 'file-replacement'),
            ('default', 'image-replacement'),
            )

        self.assertEquals(
            [set(('foo',
                  'bar'))],
            map(set, find_cyclic_dependencies(dependencies)))

    def test_indirect_cyclic_dependencies(self):
        dependencies = (
            ('foo', 'bar'),
            ('bar', 'baz'),
            ('baz', 'foo'),
            )

        self.assertEquals(
            [set(('foo',
                  'bar',
                  'baz'))],
            map(set, find_cyclic_dependencies(dependencies)))


class TestSizedGenerator(TestCase):

    def test_length(self):
        generator = SizedGenerator((i for i in range(3)), 3)
        self.assertEqual(3, len(generator))

    def test_iterating(self):
        generator = SizedGenerator((i for i in range(3)), 3)
        self.assertEqual([0, 1, 2], list(generator))


class TestSortedProfileIds(MockTestCase):

    def test_dependencies_resolved(self):
        portal_setup = self.mocker.mock()
        self.expect(portal_setup.listProfileInfo()).result([
                {'id': 'baz',
                 'dependencies': [
                        'profile-foo',
                        'profile-bar']},

                {'id': 'foo'},

                {'id': 'bar',
                 'dependencies': ['profile-foo']}]).count(1, 2)

        self.replay()

        self.assertEqual(
            ['foo', 'bar', 'baz'],
            get_sorted_profile_ids(portal_setup))

    def test_root_profiles_are_ordered_by_profile_name(self):
        """In the this example the profiles "baz" and "xy"
        have no dependencies to another and thus might be
        ordered in any order from the graph point of view.
        However, we want a cosistent ordern and therefore
        order those root nodes by name.
        """
        portal_setup = self.mocker.mock()
        self.expect(portal_setup.listProfileInfo()).result([
                {'id': 'baz',
                 'dependencies': [
                        'profile-foo']},
                {'id': 'foo'},
                {'id': 'xy'}]).count(1, 2)

        self.replay()

        self.assertEqual(
            ['foo', 'baz', 'xy'],
            get_sorted_profile_ids(portal_setup))

    def test_cyclic_dependencies(self):
        portal_setup = self.mocker.mock()
        self.expect(portal_setup.listProfileInfo()).result([
                {'id': 'foo',
                 'dependencies': ['profile-bar']},

                {'id': 'bar',
                 'dependencies': ['profile-foo']},

                {'id': 'baz',
                 'dependencies': []}]).count(1, 2)

        self.replay()

        with self.assertRaises(CyclicDependencies) as cm:
            get_sorted_profile_ids(portal_setup)

        self.assertEqual([('foo', 'bar')],
                         cm.exception.cyclic_dependencies)

        self.assertEqual([('foo', 'bar'), ('bar', 'foo')],
                         cm.exception.dependencies)


class TestFormatDuration(TestCase):

    def test_zero_seconds_is_supported(self):
        self.assertEqual('0 seconds', format_duration(0))

    def test_single_second_is_singular(self):
        self.assertEqual('1 second', format_duration(1))

    def test_multiple_seconds_is_plural(self):
        self.assertEqual('2 seconds', format_duration(2))

    def test_single_minute_is_singular(self):
        self.assertEqual(['1 minute',
                          '1 minute, 1 second',
                          '1 minute, 2 seconds'],

                         [format_duration(60),
                          format_duration(60 + 1),
                          format_duration(60 + 2)])

    def test_multiple_minutes_is_plural(self):
        self.assertEqual(['2 minutes, 1 second',
                          '2 minutes, 2 seconds'],

                         [format_duration((2 * 60) + 1),
                          format_duration((2 * 60) + 2)])

    def test_single_hour_is_singular(self):
        self.assertEqual(['1 hour',
                          '1 hour, 1 minute',
                          '1 hour, 2 minutes, 1 second',
                          '1 hour, 2 minutes, 2 seconds'],

                         [format_duration((60 * 60)),
                          format_duration((60 * 60) + 60),
                          format_duration((60 * 60) + (2 * 60) + 1),
                          format_duration((60 * 60) + (2 * 60) + 2)])

    def test_multiple_hours_is_plural(self):
        self.assertEqual(['2 hours',
                          '2 hours, 1 minute',
                          '2 hours, 2 minutes, 1 second',
                          '2 hours, 2 minutes, 2 seconds'],

                         [format_duration((2 * 60 * 60)),
                          format_duration((2 * 60 * 60) + 60),
                          format_duration((2 * 60 * 60) + (2 * 60) + 1),
                          format_duration((2 * 60 * 60) + (2 * 60) + 2)])

    def test_floating_point_seconds_are_ceiled(self):
        self.assertEqual(['1 second',
                          '1 second',
                          '2 seconds',
                          '2 seconds'],

                         [format_duration(0.1),
                          format_duration(0.9),
                          format_duration(1.1),
                          format_duration(1.9)])


class TestSubjectFromDocstring(TestCase):

    def test_one_line_only(self):
        self.assertEquals(
            'This is the subject.',
            subject_from_docstring('This is the subject.'))

    def test_whitespace_is_stripped(self):
        self.assertEquals(
            'This is the subject.',
            subject_from_docstring('\nThis is the subject.\n'))

    def test_only_subject_is_returned(self):
        self.assertEquals(
            'This is the subject.',
            subject_from_docstring('This is the subject.\n\nThis is the body.'))

    def test_multiline_subject_is_joined(self):
        self.assertEquals(
            'This is a subject, with multiple lines.',
            subject_from_docstring('This is a subject,\n'
                                   'with multiple lines.\n'
                                   '\n'
                                   'And the body.'))



from ftw.upgrade import UpgradeStep
from Products.CMFPlone.interfaces import IMigratingPloneSiteRoot
from Products.GenericSetup.interfaces import EXTENSION
from Products.GenericSetup.zcml import registerProfile
from Products.GenericSetup.zcml import upgradeStep
from zope.configuration.fields import GlobalObject
from zope.configuration.fields import Path
from zope.interface import Interface
import zope.schema


class IImportProfileUpgradeStep(Interface):
    """Register an upgrade step which imports a generic setup profile
    specific to this upgrade step.
    """

    title = zope.schema.TextLine(
        title=u"Title",
        required=True)

    description = zope.schema.TextLine(
        title=u"Upgrade step description",
        required=False)

    profile = zope.schema.TextLine(
        title=u"GenericSetup profile id",
        required=True)

    source = zope.schema.ASCII(
        title=u"Source version",
        required=True)

    destination = zope.schema.ASCII(
        title=u"Destination version",
        required=True)

    directory = Path(
        title=u'Path',
        required=True)

    handler = GlobalObject(
        title=u'Handler',
        required=False)


class DefaultUpgradeStep(UpgradeStep):
    def __call__(self):
        self.install_upgrade_profile()


def importProfileUpgradeStep(_context, title, profile, source, destination,
                             directory, description=None, handler=None):
    profile_id = "upgrade_to_%s" % destination
    registerProfile(_context, name=profile_id, title=title,
                    description=description, directory=directory,
                    provides=EXTENSION, for_=IMigratingPloneSiteRoot)

    if handler is None:
        handler = DefaultUpgradeStep

    profileid = 'profile-%s:%s' % (_context.package.__name__, profile_id)
    def handler_wrapper(portal_setup):
        return handler(portal_setup, profileid)

    upgradeStep(_context, title=title, profile=profile,
                handler=handler_wrapper, description=description,
                source=source, destination=destination)



from Acquisition import aq_base
from Products.CMFCore.utils import getToolByName


def update_security_for(obj, reindex_security=True):
    """Update the object security and reindex the security indexes in
    the catalog.
    """

    wftool = getToolByName(obj, 'portal_workflow')

    changed = False
    for permission in obj.permission_settings():
        roles_checked = [True for role in permission.get('roles', ())
                         if role.get('checked')]

        if roles_checked or not permission.get('acquire'):
            obj.manage_permission(permission['name'], roles=[],
                                  acquire=True)
            changed = True

    for workflow_id in wftool.getChainFor(obj):
        workflow = wftool.get(workflow_id)
        if not hasattr(aq_base(workflow), 'updateRoleMappingsFor'):
            continue

        if workflow.updateRoleMappingsFor(obj):
            changed = True

    if changed and reindex_security:
        obj.reindexObjectSecurity()

    return changed
